@@ -5,7 +5,7 @@ use strict;
 use warnings;
 use base qw( Exporter );
 
-our $VERSION = '0.04';
+our $VERSION = '0.06';
 
 use AnyEvent::RipeRedis;
 use AnyEvent::RipeRedis::Error;
@@ -19,7 +19,7 @@ our %ERROR_CODES;
 
 BEGIN {
   %ERROR_CODES = %AnyEvent::RipeRedis::Error::ERROR_CODES;
-  our @EXPORT_OK   = keys %ERROR_CODES;
+  our @EXPORT_OK   = ( keys %ERROR_CODES, qw( crc16 hash_slot ) );
   our %EXPORT_TAGS = ( err_codes => \@EXPORT_OK, );
 }
 
@@ -71,7 +71,20 @@ my @CRC16_TAB = (
   0x6e17, 0x7e36, 0x4e55, 0x5e74, 0x2e93, 0x3eb2, 0x0ed1, 0x1ef0,
 );
 
-my %SUB_COMMANDS = (
+my %PREDEFINED_CMDS = (
+  sort        => { readonly => 0, key_pos => 1 },
+  zunionstore => { readonly => 0, key_pos => 1 },
+  zinterstore => { readonly => 0, key_pos => 1 },
+  eval        => { readonly => 0, key_pos => 3 },
+  evalsha     => { readonly => 0, key_pos => 3 },
+  multi       => { readonly => 0, key_pos => 1, force_slot => 1 },
+  watch       => { readonly => 0, key_pos => 1, force_slot => 1 },
+  exec        => { readonly => 0, key_pos => 0, unforce_slot => 1 },
+  discard     => { readonly => 0, key_pos => 0, unforce_slot => 1 },
+  unwatch     => { readonly => 0, key_pos => 0, unforce_slot => 1 },
+);
+
+my %SUB_CMDS = (
   subscribe  => 1,
   psubscribe => 1,
 );
@@ -108,6 +121,7 @@ sub new {
   foreach my $name ( qw( password utf8 connection_timeout read_timeout
       reconnect min_reconnect_interval handle_params ) )
   {
+    next unless defined $params{$name};
     $node_params{$name} = $params{$name};
   }
   $self->{_node_params} = \%node_params;
@@ -201,6 +215,20 @@ sub crc16 {
   return $crc;
 }
 
+sub hash_slot {
+  my $key = shift;
+
+  my $tag = $key;
+
+  if ( $key =~ m/\{([^}]*?)\}/ ) {
+    if ( length $1 > 0 ) {
+      $tag = $1;
+    }
+  }
+
+  return crc16($tag) % MAX_SLOTS;
+}
+
 sub _init {
   my $self  = shift;
 
@@ -402,33 +430,25 @@ sub _discover_commands {
           return;
         }
 
-        my %commands;
+        my %commands = %PREDEFINED_CMDS;
 
         foreach my $cmd_raw ( @{$commands_raw} ) {
-          my $kwd     = lc( $cmd_raw->[0] );
-          my $flags   = $cmd_raw->[2];
-          my $key_pos = $cmd_raw->[3];
+          my $kwd = lc( $cmd_raw->[0] );
 
-          my $cmd_info = {
-            readonly    => 0,
-            movablekeys => 0,
-            key         => $key_pos,
-          };
+          next if exists $commands{$kwd};
 
-          foreach my $flag ( @{$flags} ) {
+          my $readonly = 0;
+          foreach my $flag ( @{ $cmd_raw->[2] } ) {
             if ( $flag eq 'readonly' ) {
-              $cmd_info->{readonly} = 1;
-            }
-            if ( $flag eq 'movablekeys' ) {
-              $cmd_info->{movablekeys} = 1;
+              $readonly = 1;
+              last;
             }
           }
 
-          $commands{$kwd} = $cmd_info;
-        }
-        $commands{watch}{forcing_slot} = 1;
-        foreach my $kwd ( qw( exec discard unwatch ) ) {
-          $commands{$kwd}{unforcing_slot} = 1;
+          $commands{$kwd} = {
+            readonly => $readonly,
+            key_pos  => $cmd_raw->[3],
+          };
         }
 
         $self->{_commands} = \%commands;
@@ -517,7 +537,7 @@ sub _prepare {
   else {
     $cbs = {};
     if ( ref( $args->[-1] ) eq 'CODE' ) {
-      if ( exists $SUB_COMMANDS{$cmd_name} ) {
+      if ( exists $SUB_CMDS{$cmd_name} ) {
         $cbs->{on_message} = pop @{$args};
       }
       else {
@@ -557,6 +577,12 @@ sub _route {
   my $self = shift;
   my $cmd  = shift;
 
+  if ( $cmd->{name} eq 'multi'
+    && !defined $cmd->{args}[0] )
+  {
+    croak 'Hash tag for "multi" command not specified';
+  }
+
   unless ( $self->{_ready} ) {
     if ( $self->{_init_state} == S_NEED_DO ) {
       $self->_init;
@@ -567,53 +593,67 @@ sub _route {
     return;
   }
 
-  if ( $cmd->{name} eq 'multi'
-    && !defined $self->{_forced_slot} )
-  {
-    $self->{_deferred_multi} = $cmd;
-    return;
+  my $cmd_info = $self->{_commands}{ $cmd->{kwds}[0] };
+
+  my $slot;
+  my $allow_slaves;
+
+  if ( defined $self->{_forced_slot} ) {
+    $slot = $self->{_forced_slot};
+    $cmd->{is_forced_slot} = 1;
+
+    if ( defined $cmd_info && $cmd_info->{unforce_slot} ) {
+      undef $self->{_forced_slot};
+    }
+  }
+  else {
+    my $key;
+
+    if ( defined $cmd_info ) {
+      my @tokens = ( @{ $cmd->{kwds} }, @{ $cmd->{args} } );
+      $key = $tokens[ $cmd_info->{key_pos} ];
+    }
+
+    $allow_slaves = $self->{allow_slaves};
+
+    if ( defined $key ) {
+      $slot = hash_slot($key);
+
+      if ( $cmd_info->{force_slot}
+        && !defined $self->{_forced_slot} )
+      {
+        $self->{_forced_slot} = $slot;
+      }
+
+      $allow_slaves &&= $cmd_info->{readonly};
+    }
   }
 
-  weaken($self);
+  my @nodes;
 
-  $self->_get_route( $cmd,
-    sub {
-      my $slot         = shift;
-      my $allow_slaves = shift;
-
-      my @nodes;
-
-      if ( defined $slot ) {
-        if ( defined $self->{_deferred_multi} ) {
-          my $multi = $self->{_deferred_multi};
-          undef $self->{_deferred_multi};
-          $self->_route($multi);
-
-          $self->_route($cmd);
-
-          return;
-        }
-
-        my ($range) = bsearch {
-          $slot > $_->[1] ? -1 : $slot < $_->[0] ? 1 : 0;
-        }
-        @{ $self->{_slots} };
-
-        @nodes
-            = $allow_slaves
-            ? @{ $range->[2] }
-            : ( $range->[2][0] );
-      }
-      else {
-        @nodes
-            = $allow_slaves
-            ? keys %{ $self->{_nodes_pool} }
-            : @{ $self->{_masters} };
-      }
-
-      $self->_execute( $cmd, @nodes );
+  if ( defined $slot ) {
+    my ($range) = bsearch {
+      $slot > $_->[1] ? -1 : $slot < $_->[0] ? 1 : 0;
     }
-  );
+    @{ $self->{_slots} };
+
+    @nodes
+        = $allow_slaves
+        ? @{ $range->[2] }
+        : ( $range->[2][0] );
+  }
+  else {
+    @nodes
+        = $allow_slaves
+        ? keys %{ $self->{_nodes_pool} }
+        : @{ $self->{_masters} };
+  }
+
+  if ( $cmd->{name} eq 'multi' ) {
+    $cmd->{args} = [];
+  }
+
+  $self->_execute( $cmd, @nodes );
 
   return;
 }
@@ -635,56 +675,55 @@ sub _execute {
 
   weaken($self);
 
-  my $on_reply = sub {
-    my $reply = shift;
-    my $err   = shift;
-
-    if ( defined $err ) {
-      my $err_code   = $err->code;
-      my $nodes_pool = $self->{_nodes_pool};
-
-      if ( $err_code == E_MOVED || $err_code == E_ASK ) {
-        if ( $err_code == E_MOVED ) {
-          $self->{_init_state} = S_NEED_DO;
-          $self->{_ready}      = 0;
-        }
-
-        my ($fwd_hostport) = ( split( m/\s+/, $err->message ) )[2];
-
-        unless ( defined $nodes_pool->{$fwd_hostport} ) {
-          my ( $host, $port ) = parse_hostport($fwd_hostport);
-          $nodes_pool->{$fwd_hostport} = $self->_new_node( $host, $port );
-        }
-
-        $self->_execute( $cmd, $fwd_hostport );
-
-        return;
-      }
-
-      my $on_node_error = $cmd->{on_node_error} || $self->{on_node_error};
-
-      if ( defined $on_node_error ) {
-        my $node = $nodes_pool->{$hostport};
-        $on_node_error->( $err, $node->host, $node->port );
-      }
-
-      if ( $err_code != E_CONN_CLOSED_BY_CLIENT && @nodes ) {
-        $self->_execute( $cmd, @nodes );
-        return;
-      }
-
-      $cmd->{on_reply}->( $reply, $err );
-
-      return;
-    }
-
-    $cmd->{on_reply}->($reply);
-
-    return;
-  };
-
   $node->execute( $cmd->{name}, @{ $cmd->{args} },
-    { on_reply => $on_reply,
+    { on_reply => sub {
+        my $reply = shift;
+        my $err   = shift;
+
+        if ( defined $err ) {
+          my $err_code   = $err->code;
+          my $nodes_pool = $self->{_nodes_pool};
+
+          if ( ( $err_code == E_MOVED || $err_code == E_ASK )
+            && !$cmd->{is_forced_slot} )
+          {
+            if ( $err_code == E_MOVED ) {
+              $self->{_init_state} = S_NEED_DO;
+              $self->{_ready}      = 0;
+            }
+
+            my ($fwd_hostport) = ( split( m/\s+/, $err->message ) )[2];
+
+            unless ( defined $nodes_pool->{$fwd_hostport} ) {
+              my ( $host, $port ) = parse_hostport($fwd_hostport);
+              $nodes_pool->{$fwd_hostport} = $self->_new_node( $host, $port );
+            }
+
+            $self->_execute( $cmd, $fwd_hostport );
+
+            return;
+          }
+
+          my $on_node_error = $cmd->{on_node_error} || $self->{on_node_error};
+          if ( defined $on_node_error ) {
+            my $node = $nodes_pool->{$hostport};
+            $on_node_error->( $err, $node->host, $node->port );
+          }
+
+          if ( $err_code != E_CONN_CLOSED_BY_CLIENT && @nodes ) {
+            $self->_execute( $cmd, @nodes );
+            return;
+          }
+
+          $cmd->{on_reply}->( $reply, $err );
+
+          return;
+        }
+
+        $cmd->{on_reply}->($reply);
+
+        return;
+      },
 
       defined $cmd->{on_message}
       ? ( on_message => $cmd->{on_message} )
@@ -695,107 +734,6 @@ sub _execute {
   return;
 }
 
-sub _get_route {
-  my $self = shift;
-  my $cmd  = shift;
-  my $cb   = shift;
-
-  my $cmd_info = $self->{_commands}{ $cmd->{kwds}[0] };
-
-  if ( defined $self->{_forced_slot} ) {
-    my $slot = $self->{_forced_slot};
-
-    if ( defined $cmd_info && $cmd_info->{unforcing_slot} ) {
-      undef $self->{_forced_slot};
-    }
-
-    $cb->($slot);
-
-    return;
-  }
-
-  $self->_get_key( $cmd,
-    sub {
-      my $key = shift;
-
-      unless ( defined $key ) {
-        $cb->( undef, $self->{allow_slaves} );
-        return;
-      }
-
-      my $slot = _get_slot($key);
-
-      if ( ( defined $cmd_info && $cmd_info->{forcing_slot} )
-        || defined $self->{_deferred_multi} )
-      {
-        $self->{_forced_slot} = $slot;
-      }
-
-      $cb->( $slot, $self->{allow_slaves} && $cmd_info->{readonly} );
-    }
-  );
-
-  return;
-}
-
-sub _get_key {
-  my $self = shift;
-  my $cmd  = shift;
-  my $cb   = shift;
-
-  my $cmd_info = $self->{_commands}{ $cmd->{kwds}[0] };
-
-  unless ( defined $cmd_info ) {
-    $cb->();
-    return;
-  }
-
-  my @args = ( @{ $cmd->{kwds} }, @{ $cmd->{args} } );
-
-  if ( $cmd_info->{movablekeys} ) {
-    my @nodes
-        = $self->{allow_slaves}
-        ? keys %{ $self->{_nodes_pool} }
-        : @{ $self->{_masters} };
-
-    $self->_execute(
-      { name          => 'command_getkeys',
-        args          => [@args],
-        on_node_error => $cmd->{on_node_error},
-
-        on_reply => sub {
-          my $keys = shift;
-          my $err  = shift;
-
-          if ( defined $err ) {
-            $cmd->{on_reply}->( undef, $err );
-            return;
-          }
-
-          if ( @{$keys} ) {
-            $cb->( $keys->[0] );
-          }
-          else {
-            $cb->();
-          }
-        }
-      },
-      @nodes,
-    );
-
-    return;
-  }
-
-  if ( $cmd_info->{key} > 0 ) {
-    $cb->( $args[ $cmd_info->{key} ] );
-  }
-  else {
-    $cb->();
-  }
-
-  return;
-}
-
 sub _process_input_queue {
   my $self = shift;
 
@@ -812,15 +750,14 @@ sub _process_input_queue {
 sub _reset_internals {
   my $self = shift;
 
-  $self->{_nodes_pool}     = {};
-  $self->{_masters}        = undef;
-  $self->{_slots}          = undef;
-  $self->{_commands}       = undef;
-  $self->{_init_state}     = S_NEED_DO;
-  $self->{_refresh_timer}  = undef;
-  $self->{_ready}          = 0;
-  $self->{_deferred_multi} = undef;
-  $self->{_forced_slot}    = undef;
+  $self->{_nodes_pool}    = {};
+  $self->{_masters}       = undef;
+  $self->{_slots}         = undef;
+  $self->{_commands}      = undef;
+  $self->{_init_state}    = S_NEED_DO;
+  $self->{_refresh_timer} = undef;
+  $self->{_ready}         = 0;
+  $self->{_forced_slot}   = undef;
 
   return;
 }
@@ -865,20 +802,6 @@ sub _queued_commands {
   );
 }
 
-sub _get_slot {
-  my $key = shift;
-
-  my $tag = $key;
-
-  if ( $key =~ m/\{([^}]*?)\}/ ) {
-    if ( length $1 > 0 ) {
-      $tag = $1;
-    }
-  }
-
-  return crc16($tag) % MAX_SLOTS;
-}
-
 sub _new_error {
   return AnyEvent::RipeRedis::Error->new(@_);
 }
@@ -1069,9 +992,9 @@ By default the client use kernel's connection timeout.
 =item read_timeout => $fractional_seconds
 
 Specifies read timeout. If the client could not receive a reply from the node
-after specified timeout, the client close connection and call the C<on_error>
-callback with the C<E_READ_TIMEDOUT> error. The timeout is specifies in seconds
-and can contain a fractional part.
+after specified timeout, the client close connection and call the
+C<on_node_error> callback with the C<E_READ_TIMEDOUT> error. The timeout is
+specifies in seconds and can contain a fractional part.
 
   read_timeout => 3.5,
 
@@ -1275,6 +1198,35 @@ convenient.
     }
   );
 
+=head1 TRANSACTIONS
+
+The client sends all commands of the transaction to the one master node,
+which will be selected by the hash tag in C<MULTI> command or by the first key
+in the C<WATCH> command. The hash tag of the C<MULTI> command is an extension
+of the client and it is not transferred to the Redis server.
+
+  $cluster->multi('foo');
+  $cluster->set( '{foo}bar', "some\r\nstring" );
+  $cluster->set( '{foo}car', 42 );
+  $cluster->exec(
+    sub {
+      my $reply = shift;
+      my $err   = shift;
+
+      if ( defined $err ) {
+        # error handling...
+
+        return;
+      }
+
+      # reply handling...
+    }
+  );
+
+The detailed information about the Redis transactions can be found in
+documentation on L<AnyEvent::RipeRedis> and here:
+L<http://redis.io/topics/transactions>.
+
 =head1 ERROR CODES
 
 Every error object, passed to callback, contain error code, which can be used
@@ -1311,10 +1263,18 @@ Get or set the C<on_error> callback.
 
 =head1 SERVICE FUNCTIONS
 
+Service functions provided by AnyEvent::RipeRedis::Cluster can be imported.
+
+  use AnyEvent::RipeRedis::Cluster qw( crc16 hash_slot );
+
 =head2 crc16( $data )
 
 Compute CRC16 for the specified data as defined in Redis Cluster specification.
 
+=head2 hash_slot( $key );
+
+Returns slot number by the key.
+
 =head1 SEE ALSO
 
 L<AnyEvent::RipeRedis>
